"""Base test case support for Sphinx extensions."""

import os
import shutil
import tempfile
from contextlib import contextmanager
from unittest import TestCase

from sphinx_testing.util import TestApp, docutils_namespace

import beanbag_docutils


_theme_path = os.path.abspath(os.path.join(beanbag_docutils.__file__,
                                           '..', '..', 'tests', 'sphinx',
                                           'themes'))


class SphinxExtTestCase(TestCase):
    """Base class for test cases for Sphinx extensions.

    This provides common functionality for setting up and rendering a ReST
    document to a content-only HTML string.
    """

    #: Class-wide configuration for the renders.
    config = {}

    #: A list of extensions required for the tests.
    extensions = []

    maxDiff = 1000000

    @contextmanager
    def with_sphinx_env(self, config={}, builder_name='html'):
        """Run within a Sphinx environment.

        Version Added:
            1.10

        Args:
            config (dict, optional):
                Additional configuration to set for phinx.

        Context:
            dict:
            A dictionary containing:

            Keys:
                app (sphinx.application.Sphinx):
                    The Sphinx application that was set up.

                config (sphinx.config.Config):
                    The Sphinx configuration object.

                srcdir (unicode):
                    The location to place ReST source files in.
        """
        new_config = {
            'extensions': self.extensions,
            'html_theme': 'test',
            'html_theme_path': [_theme_path],
        }
        new_config.update(self.config)
        new_config.update(config)

        with docutils_namespace():
            app = TestApp(buildername=builder_name,
                          create_new_srcdir=True,
                          copy_srcdir_to_tmpdir=False,
                          confoverrides=new_config)

            try:
                yield {
                    'app': app,
                    'config': app.config,
                    'srcdir': app.srcdir,
                }
            finally:
                app.cleanup()

    def render_doc(self, doc_content, config={}, builder_name='html',
                   extra_files={}):
        """Render a ReST document to a string.

        This will set up a Sphinx environment, based on the extensions and
        configuration provided by the consumer, and perform a render of the
        given ReST content. The resulting string will be a rendered version
        of just that content.

        Args:
            doc_content (unicode):
                The ReST content to render.

            config (dict, optional):
                Additional configuration to set for the render.

        Returns:
            unicode:
            The rendered content.
        """
        if builder_name == 'html':
            out_filename = 'contents.html'
        else:
            raise ValueError('"%s" is not a supported builder name'
                             % builder_name)

        with self.with_sphinx_env(config=config,
                                  builder_name=builder_name) as ctx:
            srcdir = ctx['srcdir']

            old_cwd = os.getcwd()
            os.chdir(srcdir)

            try:
                with open('contents.rst', 'w') as fp:
                    fp.write(doc_content)

                for path, contents in extra_files.items():
                    dirname = os.path.dirname(path)

                    if not os.path.exists(dirname):
                        os.makedirs(dirname)

                    with open(path, 'wb') as fp:
                        fp.write(contents)

                app = ctx['app']
                app.build()

                with open(app.outdir / out_filename, 'r') as fp:
                    return fp.read().strip()
            finally:
                os.chdir(old_cwd)



"""Sphinx extension for srcsets in images.

.. versionadded:: 2.1


This extension adds a ``sources`` option to the standard image directives,
enabling responsive image support via srcsets.

These are specified a bit differently from ``<img srcset="...">`` values.
The descriptor goes first, and a comma between entries is optional (a blank
line can be used instead). For example:

.. code-block:: rst

   .. image:: path/to/file.png
      :sources: 2x path/to/file@2x.png
                3x path/to/file@3x.png
                100w path/to/file@100w.png
                200h path/to/file@200h.png

If ``sources`` is not explicitly provided, but files with those standard
``@``-based suffixes exist alongside the referenced main image, they'll
automatically be used to define the srcsets of the image. The ``1x`` entry is
also automatically inserted based on the main image.

If relying on the default of scanning for srcset images, this becomes a
zero-configuration, drop-in solution for all Sphinx documentation codebases.


Setup
=====

To use this, you just need to add the extension in :file:`conf.py`::

    extensions = [
        ...
        'beanbag_docutils.sphinx.ext.image_srcsets',
        ...
    ]
"""

import os
import posixpath
import re
from collections import OrderedDict
from glob import glob
from typing import Dict

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives.images import Image
from six.moves.urllib.parse import quote as urllib_quote
from sphinx.application import Sphinx
from sphinx.util.i18n import search_image_for_language


def _get_srcsets(
    node  # type: nodes.image
):  # type: (...) -> Dict[str, str]
    """Return a normalized version of all srcsets for an image node.

    This will convert a ``sources`` option to a dictionary mapping descriptors
    (such as ``2x``, ``100w``, etc.) to URLs.

    These will be cached for future lookup.

    Args:
        node (docutils.nodes.image):
            The image node to retrieve sources from.

    Returns:
        dict:
        The mapping of descriptors to URLs.
    """
    try:
        norm_srcsets = node.attributes['_srcsets']
    except KeyError:
        srcset = node.attributes.get('sources')
        norm_srcsets = OrderedDict()

        if srcset:
            norm_srcsets['1x'] = node.attributes['uri']

            for source in re.split(r',|\n+', srcset):
                source = source.strip()

                if source:
                    descriptor, url = source.split(' ', 1)
                    norm_srcsets[descriptor.strip()] = url.strip()

        node.attributes['_srcsets'] = norm_srcsets

    return norm_srcsets


def _visit_image_html(
    self,
    node   # type: nodes.image
):  # type: (...) -> None
    """Process an Image node.

    This will update the HTML of the image node with a ``srcsets=`` attribute
    if srcsets are needed.

    Args:
        node (docutils.nodes.image):
            The image node to process.
    """
    # Use the default logic to build the image tag, since it's non-trivial.
    type(self).visit_image(self, node)

    images = self.builder.env.images
    base_images_path = self.builder.imgpath
    srcsets = _get_srcsets(node)

    if srcsets:
        last_tag = self.body[-1]
        assert last_tag.startswith('<img ')

        self.body[-1] = (
            '<img srcset="%s" %s' % (
                ', '.join(
                    '%s %s' % (
                        posixpath.join(base_images_path,
                                       urllib_quote(images[url][1])),
                        source)
                    for source, url in srcsets.items()
                ),
                last_tag[len('<img '):],
            )
        )


def collect_srcsets(
    app,     # type: Sphinx
    doctree  # type: nodes.document
):  # type: (...) -> None
    """Collect all images referenced by image nodes or scanned in directories.

    This will collect any explicit values defined via our ``sources`` option
    for image directives. If ``sources`` is not specified, but there are files
    in the directory with ``@2x``, ``@3x``, ``@100w`` ``@100h``, etc.
    descriptors, those will be collected instead and associated with the image.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application being run.

        doctree (docutils.nodes.document):
            The document tree being processed.
    """
    env = app.env
    images = env.images
    docname = env.docname

    if hasattr(doctree, 'findall'):
        # This is the modern way of finding nodes.
        findall = doctree.findall
    else:
        # This is pending deprecation in docutils.
        findall = doctree.traverse

    for node in findall(nodes.image):
        srcsets = _get_srcsets(node)

        if not srcsets:
            image_path = search_image_for_language(node['uri'], env)
            rel_image_path, abs_image_path = env.relfn2path(
                search_image_for_language(node['uri'], env),
                docname)
            base_filename, ext = os.path.splitext(image_path)
            candidates = glob('%s@*%s' % (base_filename, ext))

            if candidates:
                srcsets['1x'] = node['uri']

                pattern = re.compile(r'%s@(\d+[xwh])%s'
                                     % (re.escape(base_filename),
                                        re.escape(ext)))

                for candidate in sorted(candidates):
                    m = pattern.match(candidate)

                    if m:
                        descriptor = m.group(1)

                        if descriptor not in srcsets:
                            srcsets[descriptor] = candidate

        for descriptor, image_path in list(srcsets.items()):
            image_path = env.relfn2path(
                search_image_for_language(image_path, env),
                docname)[0]

            env.dependencies[docname].add(image_path)
            images.add_file(docname, image_path)


def setup(
    app  # type: Sphinx
):  # type: (...) -> None
    """Set up the Sphinx extension.

    This listens for the events needed to collect and bundle images for
    srcsets, and update the resulting HTML to specify them.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application being run.
    """
    Image.option_spec['sources'] = directives.unchanged

    app.add_node(nodes.image,
                 html=(_visit_image_html, None),
                 override=True)

    app.connect('doctree-read', collect_srcsets)



"""Sphinx extension to add useful Django-related functionality.

This adds some improvements when working with Django-based classes in autodocs,
and when referencing Django documentation.

Localized strings using :py:func:`~django.utils.translation.ugettext_lazy` will
be turned into actual strings for the documentation, which is useful for forms
and models.

Roles are also added for common cross-references.


Setup
=====

To use this, you just need to add the extension in :file:`conf.py`::

    extensions = [
        ...
        'beanbag_docutils.sphinx.ext.django_utils',
        ...
    ]


Roles
=====

.. rst:role:: setting

   Creates and links to references for Django settings (those that live in
   ``django.conf.settings``).

   For example:

   .. code-block:: rst

       .. setting:: MY_SETTING

       Settings go here.

       And then to reference it: :setting:`MY_SETTING`.
"""

from django.utils.functional import Promise


def _repr_promise(promise):
    """Return a sane representation of a lazy localized string.

    If the promise is a result of ugettext_lazy(), it will be converted into
    a Unicode string before generating a representation.
    """
    if hasattr(promise, '_proxy____text_cast'):
        return '_(%r)' % str(promise)

    return super(promise.__class__, promise).__repr__(promise)


def setup(app):
    """Set up the Sphinx extension.

    This registers cross-references and fixes up the lazily-localized strings
    for display.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application to listen to events on.
    """
    Promise.__repr__ = _repr_promise

    app.add_crossref_type(directivename=str('setting'),
                          rolename=str('setting'),
                          indextemplate=str('pair: %s; setting'))



#: The version of beanbag_docutils.
#:
#: This is in the format of:
#:
#:   (Major, Minor, Micro, alpha/beta/rc/final, Release Number, Released)
#:
VERSION = (2, 1, 0, 'final', 0, True)


def get_version_string():
    """Return the version as a human-readable string.

    Returns:
        unicode:
        The version number as a human-readable string.
    """
    version = '%s.%s' % (VERSION[0], VERSION[1])

    if VERSION[2]:
        version += ".%s" % VERSION[2]

    if VERSION[3] != 'final':
        if VERSION[3] == 'rc':
            version += ' RC%s' % VERSION[4]
        else:
            version += ' %s %s' % (VERSION[3], VERSION[4])

    if not is_release():
        version += " (dev)"

    return version


def get_package_version():
    """Return the version as a Python package version string.

    Returns:
        unicode:
        The version number as used in a Python package.
    """
    version = '%s.%s' % (VERSION[0], VERSION[1])

    if VERSION[2]:
        version += ".%s" % VERSION[2]

    if VERSION[3] != 'final':
        version += '%s%s' % (VERSION[3], VERSION[4])

    return version


def is_release():
    """Return whether this is a released version.

    Returns:
        bool:
        ``True`` if this is a released version of the package.
    """
    return VERSION[5]


#: An alias for the the version information from :py:data:`VERSION`.
#:
#: This does not include the last entry in the tuple (the released state).
__version_info__ = VERSION[:-1]

#: An alias for the version used for the Python package.
__version__ = get_package_version()



"""Sphinx extension to link to source code on GitHub.

This links to source code for modules, classes, etc. to the correct line on
GitHub. This prevents having to bundle the source code along with the
documentation, and better ties everything together.


Setup
=====

To use this, you'll need to import the module and define your own
:py:func:`linkcode_resolve` in your :file:`conf.py`::

    from beanbag_docutils.sphinx.ext.github import github_linkcode_resolve

    extensions = [
        ...
        'sphinx.ext.linkcode',
        ...
    ]

    def linkcode_resolve(domain, info):
        return github_linkcode_resolve(
            domain=domain,
            info=info,
            allowed_module_names=['mymodule'],
            github_org_id='myorg',
            github_repo_id='myrepo',
            branch='master',
            source_prefix='src/')

``source_prefix`` and ``allowed_module_names`` are optional. See the
docs for :py:func:`github_linkcode_resolve` for more information.
"""

import ast
import inspect
import logging
import re
import subprocess
import sys

logger = logging.getLogger(__name__)


GIT_BRANCH_CONTAINS_RE = re.compile(r'^\s*([^\s]+)\s+([0-9a-f]+)\s.*')

# We'll store 5 items in the AST cache by default. This will ensure that we
# don't re-parse the same tree any more often than we have to, and leave
# room for some parallel processing if needed.
_AST_CACHE_MAX_SIZE = 5


_head_ref = None
_ast_cache = []


def _run_git(cmd):
    """Run git with the given arguments, returning the output.

    Args:
        cmd (list of unicode):
            A list of arguments to pass to :command:`git`.

    Returns:
        bytes:
        The resulting output from the command.

    Raises:
        subprocess.CalledProcessError:
            Error calling into git.
    """
    assert cmd
    assert all(cmd)

    p = subprocess.Popen(['git'] + cmd, stdout=subprocess.PIPE)
    output, error = p.communicate()
    ret_code = p.poll()

    if ret_code:
        raise subprocess.CalledProcessError(ret_code, 'git')

    assert isinstance(output, bytes)

    return output


def _git_get_nearest_tracking_branch(merge_base, remote='origin'):
    """Return the nearest tracking branch for the given Git repository.

    Args:
        merge_base (unicode):
            The merge base used to locate the nearest tracking branch.

        remote (origin, optional):
            The remote name.

    Returns:
        unicode:
        The nearest tracking branch, or ``None`` if not found.
    """
    try:
        _run_git(['fetch', 'origin', '%s:%s' % (merge_base, merge_base)])
    except Exception:
        # Ignore, as we may already have this. Hopefully it won't fail later.
        pass

    lines = _run_git(['branch', '-rv', '--contains', merge_base]).splitlines()

    remote_prefix = '%s/' % remote
    best_distance = None
    best_ref_name = None

    for line in lines:
        m = GIT_BRANCH_CONTAINS_RE.match(line.decode('utf-8').strip())

        if m:
            ref_name = m.group(1)
            sha = m.group(2)

            if (ref_name.startswith(remote_prefix) and
                not ref_name.endswith('/HEAD')):

                distance = len(_run_git(['log',
                                         '--pretty=format:%H',
                                         '...%s' % sha]).splitlines())

                if best_distance is None or distance < best_distance:
                    best_distance = distance
                    best_ref_name = ref_name

    if best_ref_name and best_ref_name.startswith(remote_prefix):
        # Strip away the remote.
        best_ref_name = best_ref_name[len(remote_prefix):]

    return best_ref_name


def _get_git_doc_ref(branch):
    """Return the commit SHA used for linking to source code on GitHub.

    The commit SHA will be cached for future lookups.

    Args:
        branch (unicode):
            The branch to use as a merge base.

    Returns:
        unicode:
        The commit SHA used for any links, if found, or ``None`` if not.
    """
    global _head_ref

    if not _head_ref:
        _head_ref = None

        try:
            tracking_branch = _git_get_nearest_tracking_branch(branch)

            if tracking_branch:
                _head_ref = (
                    _run_git(['rev-parse', tracking_branch]).strip()
                    .decode('utf-8')
                )
        except subprocess.CalledProcessError:
            pass

    return _head_ref


def _find_path_in_ast_nodes(nodes, path):
    """Return an AST node for an object path, if found.

    This walks the AST tree, guided by the given identifier path, trying to
    locate the referenced identifier.

    If a node represented by the path can be found, the node will be returned.

    Version Added:
        2.0

    Args:
        nodes (list or ast.AST):
            The list of nodes or the AST object to walk.

        path (list of unicode):
            The identifier path.

    Returns:
        ast.Node:
        The node, if found, or ``None`` if not found.
    """
    name = path[0]

    if not isinstance(nodes, list):
        nodes = ast.iter_child_nodes(nodes)

    for node in nodes:
        # If this is an explicit assignment, check to see if any of the
        # targets of the assignment is the path we're looking for.
        if isinstance(node, ast.Assign):
            target_names = {
                getattr(_target, 'id', None)
                for _target in node.targets
            }

            if name in target_names:
                # We found it. We're done.
                #
                # Note that there might be more items in path, but if so,
                # then we're documenting something like a namedtuple() that
                # got code-injected. There isn't likely to be any code to
                # link to beyond this.
                return node

        # If this is anything else, see if it has the next name in the path.
        if hasattr(node, 'name') and node.name == name:
            if len(path) > 1:
                # This is a match, but we have more to process. Recurse.
                return _find_path_in_ast_nodes(node.body, path[1:])
            else:
                # This was the last part we needed. Return the node.
                return node

    return None


def _find_ast_node_for_path(module, path):
    """Return the AST node for an object path, if found.

    Version Added:
        2.0

    Args:
        module (module):
            The module in which to begin the search.

        path (unicode):
            The ``.``-delimited path to the node.

    Returns:
        ast.Node:
        The resulting node, if found, or ``None`` if not found.
    """
    global _ast_cache

    tree = None

    # See if we already have a parsed AST in cache.
    for _tree, _module in _ast_cache:
        if _module is module:
            tree = _tree
            break

    if tree is None:
        # We don't have one in cache, so build it and push the last item
        # out of cache.
        try:
            lines = inspect.findsource(module)[0]
            tree = ast.parse(''.join(lines))
            _ast_cache = (
                [(module, tree)] +
                _ast_cache[:_AST_CACHE_MAX_SIZE - 1]
            )
        except Exception as e:
            logger.exception('Failed to parse AST tree for %r: %s',
                             module, e)
            return None

    try:
        return _find_path_in_ast_nodes(tree, path)
    except Exception as e:
        logger.exception('Failed to look up object %s: %s',
                         '.'.join(path), e)
        return None


def github_linkcode_resolve(domain, info, github_org_id, github_repo_id,
                            branch, source_prefix='', allowed_module_names=[],
                            *, github_url='https://github.com'):
    """Return a link to the source on GitHub for the given autodoc info.

    This takes some basic information on the GitHub project, branch, and
    what modules are considered acceptable, and generates a link to the
    approprite line on the GitHub repository browser for the class, function,
    variable, or other object.

    Version Changed:
        2.1:
        Added ``github_host`` and ``github_scheme`` arguments.

    Args:
        domain (unicode):
            The autodoc domain being processed. This only accepts "py", and
            comes from the original :py:func:`linkcode_resolve` call.

        info (dict):
            Information on the item being linked to. This comes from the
            original :py:func:`linkcode_resolve` call.

        github_org_id (unicode):
            The GitHub organization ID.

        github_repo_id (unicode):
            The GitHub repository ID.

        branch (unicode):
            The branch used as a merge base to find the appropriate commit
            to link to. Callers may want to compute this off of the version
            number of the project, or some other information.

        source_prefix (unicode, optional):
            A prefix for any linked filename, in case the module is not at
            the top of the source tree.

        allowed_module_names (list of unicode, optional):
            The list of top-level module names considered valid for links.
            If provided, links will only be generated if residing somewhere
            in one of the provided module names.

        github_url (str, optional):
            The URL to the base of the GitHub server.

            Version Added:
                2.1

    Returns:
        str:
        The link to the source. This will be ``None`` if a URL couldn't be
        calculated.
    """
    module_name = info['module']

    if (domain != 'py' or
        not module_name or
        (allowed_module_names and
         module_name.split('.')[0] not in allowed_module_names)):
        # These aren't the modules you're looking for.
        return None

    # Grab the name of the source file.
    filename = module_name.replace('.', '/') + '.py'

    # Grab the module referenced in the docs.
    submod = sys.modules.get(module_name)

    if submod is None:
        return None

    # Split that, trying to find the module at the very tail of the module
    # path.
    node = _find_ast_node_for_path(submod, info['fullname'].split('.'))

    if node is None:
        return None

    # Build a reference for the line number in GitHub.
    linespec = '#L%d' % node.lineno

    # Get the branch/tag/commit to link to.
    ref = _get_git_doc_ref(branch)

    if not ref:
        return None

    assert isinstance(ref, str)

    github_url = github_url.rstrip('/')

    return (
        f'{github_url}/{github_org_id}/{github_repo_id}/blob/'
        f'{ref}/{source_prefix}{filename}{linespec}'
    )


def clear_github_linkcode_caches():
    """Clear the internal caches for GitHub code linking.

    This is primarily intended for unit tests.

    Version Added:
        2.0
    """
    global _head_ref, _ast_cache

    _head_ref = None
    _ast_cache = []



# NOTE: The line numbers in this file matters for tests.
from collections import namedtuple


class ClassA(object):
    pass


class ClassB(object):
    """Docstring here."""

    my_var = True

    def __init__(self):
        pass

    def do_thing(self, a, b):
        pass


def my_func():
    pass


my_var = ClassB

MyNamedTuple = namedtuple('MyNamedTuple', ('a', 'b', 'c'))



"""Sphinx extension to enhance intersphinx support.

This fixes some reference issues with :rst:role:`option` (see
https://github.com/sphinx-doc/sphinx/pull/3769 for the equivalent upstream
fix).

It also introduces a ``.. default-intersphinx::`` directive that allows for
specifying one or more intersphinx set prefixes that should be tried if a
reference could not be found. For example::

    .. default-intersphinx:: myapp1.5 python

    :ref:`some-reference`

This does affect the process by which missing references are located. If an
unprefixed reference is used, it will only match if the prefix is in the list
above, which differs from the default behavior of looking through all
intersphinx mappings.


Setup
=====

This extension must be added to ``exetnsions`` in :file:`conf.py` after the
:py:mod:`sphinx.ext.intersphinx` extension is added. For example::

    extensions = [
        ...
        'sphinx.ext.intersphinx',
        'beanbag_docutils.sphinx.ext.intersphinx',
        ...
    ]
"""

import re

from docutils.parsers.rst import Directive
from sphinx.errors import ExtensionError
from sphinx.ext import intersphinx


class DefaultIntersphinx(Directive):
    """Specifies one or more default intersphinx sets to use."""

    required_arguments = 1
    optional_arguments = 100

    SPLIT_RE = re.compile(r',\s*')

    def run(self):
        """Run the directive.

        Returns:
            list:
            An empty list, always.
        """
        env = self.state.document.settings.env
        env.metadata[env.docname]['default-intersphinx-prefixes'] = \
            self.arguments

        return []


def _on_missing_reference(app, env, node, contnode):
    """Handler for missing references.

    This will attempt to fix references to options and then attempt to
    apply default intersphinx prefixes (if needed) before resolving a
    reference using intersphinx.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application processing the document.

        env (sphinx.environment.BuildEnvironment):
            The environment for this doc build.

        node (sphinx.addnodes.pending_xref):
            The pending reference to resolve.

        contnode (docutils.nodes.literal):
            The context for the reference.

    Returns:
        list:
        The list of any reference nodes, as created by intersphinx.
    """
    orig_target = node['reftarget']
    target = orig_target
    domain = node.get('refdomain')

    # See if we're referencing a std:option. Sphinx (as of 1.6.1) does not
    # properly link these. A pull request has been opened to fix this
    # (https://github.com/sphinx-doc/sphinx/pull/3769). Until we can make
    # use of that, we're including the workaround here.
    if domain == 'std' and node['reftype'] == 'option':
        # Options are stored in the inventory as "program-name.--option".
        # In order to look up the option, the target will need to be
        # converted to this format.
        #
        # Ideally, we'd be able to utilize the same logic as found in
        # StandardDomain._resolve_option_xref, but we don't have any
        # information on the progoptions data stored there. Instead, we
        # try to determine if the target already has a program name or
        # an intersphinx doc set and normalize the contents to match the
        # option reference name format.
        i = target.rfind(' ')

        if i != -1:
            # The option target contains a program name and an option
            # name. We can easily normalize this to be in
            # <progname>.<option> format.
            target = '%s.%s' % (target[:i], target[i + 1:])
            target = target.replace(' ', '-')
        else:
            # Since a space was not found, and a program name is needed
            # to complete the reference, we'll see if a ".. program::"
            # has been set in this file. If so, we'll put that into the
            # target name (being careful to consider any intersphinx doc
            # set name that may be prefixed).
            progname = node.get('std:program')

            if progname:
                if ':' in target:
                    setname, newtarget = target.split(':', 1)
                    target = '%s:%s.%s' % (setname, progname, newtarget)
                else:
                    target = '%s.%s' % (progname, target)

    if ':' not in target:
        prefixes = \
            env.metadata[env.docname].get('default-intersphinx-prefixes')

        if prefixes:
            # Try all supported prefixes in order. These are the only allowed
            # to be inferred.
            for prefix in prefixes:
                old_content = contnode[0]
                node['reftarget'] = '%s:%s' % (prefix, target)
                result = intersphinx.missing_reference(app, env, node,
                                                       contnode)

                if result:
                    return result

                # Couldn't find it. Go back to the original target and try
                # again.
                node['reftarget'] = orig_target
                contnode[0] = old_content

            return None

    return intersphinx.missing_reference(app, env, node, contnode)


def setup(app):
    """Set up the Sphinx extension.

    This listens for the events needed to handle missing references, and
    registers directives.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application building the docs.
    """
    app.add_directive('default-intersphinx', DefaultIntersphinx)

    # Disconnect the other intersphinx listener. We're going to override it.
    listeners = app.events.listeners.get('missing-reference', {})
    intersphinx_listener_id = None

    for listener in listeners:
        if listener.handler == intersphinx.missing_reference:
            intersphinx_listener_id = listener.id
            break

    if intersphinx_listener_id is not None:
        app.events.disconnect(intersphinx_listener_id)
    else:
        raise ExtensionError(
            'beanbag_docutils.sphinx.ext.intersphinx_utils '
            'must come after sphinx.ext.intersphinx')

    app.connect('missing-reference', _on_missing_reference)



"""Unit tests for beanbag_docutils.sphinx.ext.http_role."""

from beanbag_docutils.sphinx.ext.tests.testcase import SphinxExtTestCase


class HttpRoleTests(SphinxExtTestCase):
    """Unit tests for HTTP roles."""

    extensions = [
        'beanbag_docutils.sphinx.ext.http_role',
    ]

    def test_http_status_codes_format(self):
        """Testing http-status-codes-format directive with :http:`...` role"""
        self.assertEqual(
            self.render_doc(
                '.. http-status-codes-format:: HTTP Code %(code)s\n'
                '\n'
                'This is :http:`404`.'
            ),
            (
                '<p>This is <a class="reference external"'
                ' href="https://wikipedia.org/wiki/List_of_HTTP_status_codes'
                '#404">HTTP Code 404</a>.</p>'
            )
        )

    def test_http_role(self):
        """Testing :http:`...` role"""
        self.assertEqual(
            self.render_doc('This is :http:`404`.'),
            (
                '<p>This is <a class="reference external"'
                ' href="https://wikipedia.org/wiki/List_of_HTTP_status_codes'
                '#404">HTTP 404 Not Found</a>.</p>'
            )
        )

    def test_http_role_with_http_status_codes_format(self):
        """Testing :http:`...` role with custom http_status_codes_format"""
        self.assertEqual(
            self.render_doc(
                'This is :http:`404`.',
                config={
                    'http_status_codes_format': 'HTTP Code %(code)s',
                }),
            (
                '<p>This is <a class="reference external"'
                ' href="https://wikipedia.org/wiki/List_of_HTTP_status_codes'
                '#404">HTTP Code 404</a>.</p>'
            )
        )

    def test_http_role_with_http_status_codes_url(self):
        """Testing :http:`...` role with custom http_status_codes_url"""
        self.assertEqual(
            self.render_doc(
                'This is :http:`404`.',
                config={
                    'http_status_codes_url': 'https://example.com/http/%s',
                }),
            (
                '<p>This is <a class="reference external"'
                ' href="https://example.com/http/404">HTTP 404 Not Found</a>.'
                '</p>'
            )
        )



#!/usr/bin/env python

from setuptools import setup, find_packages

from beanbag_docutils import get_package_version


PACKAGE_NAME = 'beanbag-docutils'


with open('README.rst', 'r') as fp:
    readme = fp.read()


setup(
    name=PACKAGE_NAME,
    version=get_package_version(),
    license='MIT',
    description="Sphinx utility modules for Beanbag's documentation format.",
    long_description=readme,
    url='https://github.com/beanbaginc/beanbag-docutils',
    maintainer='Christian Hammond',
    maintainer_email='christian@beanbaginc.com',
    packages=find_packages(),
    install_requires=[
        'Sphinx',
    ],
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Documentation',
        'Topic :: Documentation :: Sphinx',
        'Topic :: Software Development',
        'Topic :: Software Development :: Documentation',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



"""Unit tests for beanbag_docutils.sphinx.ext.django_utils."""

from sphinx import version_info as sphinx_version_info

from beanbag_docutils.sphinx.ext.tests.testcase import SphinxExtTestCase


class DjangoUtilsTests(SphinxExtTestCase):
    """Unit tests for Django doc utilities."""

    extensions = [
        'beanbag_docutils.sphinx.ext.django_utils',
    ]

    def test_setting_crossref(self):
        """Testing Django :setting: crossref"""
        rendered = self.render_doc(
            '.. setting:: MY_TEST\n'
            '\n'
            'Some content.\n'
            '\n'
            'Now link to :setting:`MY_TEST`.'
        )

        if sphinx_version_info[:2] >= (5, 0):
            # Sphinx >= 5.0
            self.assertEqual(
                rendered,
                (
                    '<p id="std-setting-MY_TEST">Some content.</p>\n'
                    '<p>Now link to <a class="reference internal"'
                    ' href="#std-setting-MY_TEST"><code class="xref std'
                    ' std-setting docutils literal notranslate">'
                    '<span class="pre">MY_TEST</span></code></a>.</p>'
                )
            )
        elif sphinx_version_info[:2] >= (3, 0):
            # Sphinx >= 3.0, <5.0
            self.assertEqual(
                rendered,
                (
                    '<p id="std-setting-MY_TEST">'
                    '<span id="std:setting-MY_TEST"></span>'
                    'Some content.</p>\n'
                    '<p>Now link to <a class="reference internal"'
                    ' href="#std-setting-MY_TEST"><code class="xref std'
                    ' std-setting docutils literal notranslate">'
                    '<span class="pre">MY_TEST</span></code></a>.</p>'
                )
            )
        else:
            # Sphinx < 3.0
            self.assertEqual(
                rendered,
                (
                    '<p id="std:setting-MY_TEST">Some content.</p>\n'
                    '<p>Now link to <a class="reference internal"'
                    ' href="#std:setting-MY_TEST"><code class="xref std'
                    ' std-setting docutils literal notranslate">'
                    '<span class="pre">MY_TEST</span></code></a>.</p>'
                )
            )



"""Sphinx extension to add utilities for autodoc.

This enhances autodoc support for Beanbag's docstring format and to allow for
excluding content from docs.


Beanbag's Docstrings
====================

By setting ``napoleon_beanbag_docstring = True`` in :file:`conf.py`, and
turning off ``napoleon_google_docstring``, Beanbag's docstring format can be
used.

This works just like the Google docstring format, but with a few additions:

* A new ``Context:`` section to describe what happens within the context of a
  context manager (including the variable).

* New ``Model Attributes:`` and ``Option Args:`` sections for defining the
  attributes on a model or the options in a dictionary when using JavaScript.

* New ``Deprecated:``, ``Version Added:``, and ``Version Changed:`` sections
  for defining version-related information.

* Parsing improvements to allow for wrapping argument types across lines,
  which is useful when you have long module paths that won't fit on one line.

This requires the ``sphinx.ext.napoleon`` module to be loaded.


Excluding Content
=================

A module can define top-level ``__autodoc_excludes__`` or ``__deprecated__``
lists. These are in the same format as ``__all__``, in that they take a list of
strings for top-level classes, functions, and variables. Anything listed here
will be excluded from any autodoc code.

``__autodoc_excludes__`` is particularly handy when documenting an
:file:`__init__.py` that imports contents from a submodule and re-exports it in
``__all__``. In this case, autodoc would normally output documentation both in
:file:`__init__.py` and the submodule, but that can be avoided by setting::

    __autodoc_excludes = __all__

Excludes can also be defined globally, filtered by the type of object the
docstring would belong to. See the documentation for autodoc-skip-member_ for
more information. You can configure this in :file:`conf.py` by doing::

    autodoc_excludes = {
        # Applies to modules, classes, and anything else.
        '*': [
            '__annotations__',
            '__dict__',
            '__doc__',
            '__module__',
            '__weakref__',
        ],
        'class': [
            # Useful for Django models.
            'DoesNotExist',
            'MultipleObjectsReturned',
            'objects',

            # Useful forms.
            'base_fields',
            'media',
        ],
    }

That's just an example, but a useful one for Django users.

By default, ``autodoc_excludes`` is set to::

    autodoc_excludes = {
        # Applies to modules, classes, and anything else.
        '*': [
            '__dict__',
            '__doc__',
            '__module__',
            '__weakref__',
        ],
    }

If overriding, you can set ``'__defaults__': True`` to merge your changes in
with these defaults.

.. versionchanged:: 2.0
   Changed from a default of an empty dictionary, and added the
   ``__defaults__`` option.


.. _autodoc-skip-member:
   http://www.sphinx-doc.org/en/stable/ext/autodoc.html#event-autodoc-skip-member


Setup
=====

.. versionchanged:: 2.0
   Improved setup by enabling other default extensions and options when
   enabling this extension.


To use this, add the extension in :file:`conf.py`::

    extensions = [
        ...
        'beanbag_docutils.sphinx.ext.autodoc_utils',
        ...
    ]

This will automatically enable the :py:mod:`sphinx.ext.autodoc`,
:py:mod:`sphinx.ext.intersphinx`, and :py:mod:`sphinx.ext.napoleon` extensions,
and enable the following autodoc options::

    autodoc_member_order = 'bysource'
    autoclass_content = 'class'
    autodoc_default_options = {
        'members': True,
        'special-members': True,
        'undoc-members': True,
        'show-inheritance': True,
    }

These default options can be turned off by setting
``use_beanbag_autodoc_defaults=False``.


Configuration
=============

``autodoc_excludes``
    Optional global exclusions to apply, as shown above.

``napoleon_beanbag_docstring``
    Set whether the Beanbag docstring format should be used.

    This is the default as of ``beanbag-docutils`` 2.0.

``use_beanbag_autodoc_defaults``
    Set whether autodoc defaults should be used.

    .. versionadded:: 2.0
"""

import re
import sys

from sphinx import version_info
from sphinx.ext.napoleon.docstring import GoogleDocstring

from beanbag_docutils import VERSION

try:
    from sphinx.ext.napoleon.docstring import _convert_type_spec
except ImportError:
    def _convert_type_spec(type_part, type_aliases):
        try:
            return type_aliases[type_part]
        except KeyError:
            if type_part == 'None':
                return ':obj:`None`'
            else:
                return f':class:`{type_part}`'


class BeanbagDocstring(GoogleDocstring):
    """Docstring parser for the Beanbag documentation.

    This is based on the Google docstring format used by Napoleon, but with
    a few additions:

    * Support for describing contexts in a context manager (using the
      ``Context:`` section, which works like ``Returns`` or ``Yields``).

    * ``Model Attributes:`` and ``Option Args:`` argument sections.

    * Parsing improvements for arguments to allow for wrapping across lines,
      for long module paths.
    """

    partial_typed_arg_start_re = \
        re.compile(r'\s*(.+?)\s*\(\s*(.*[^\s]+)\s*[^:)]*$')
    partial_typed_arg_end_re = re.compile(r'\s*(.+?)\s*\):$')

    extra_returns_sections = [
        ('context', 'Context', {}),
        ('type', 'Type', {
            'require_type': True,
        }),
    ]

    extra_fields_sections = [
        ('keys', 'Keys'),
        ('model attributes', 'Model Attributes'),
        ('option args', 'Option Args'),
        ('tuple', 'Tuple'),
    ]

    extra_version_info_sections = [
        ('deprecated', 'deprecated'),
        ('version added', 'versionadded'),
        ('version changed', 'versionchanged'),
    ]

    MAX_PARTIAL_TYPED_ARG_LINES = 3

    COMMA_RE = re.compile(r',\s*')

    _USES_LINES_DEQUE = (version_info[:2] >= (5, 1))

    def __init__(self, *args, **kwargs):
        """Initialize the parser.

        Args:
            *args (tuple):
                Positional arguments for the parent.

            **kwargs (dict):
                Keyword arguments for the parent.
        """
        super(BeanbagDocstring, self).__init__(*args, **kwargs)

        for keyword, label, options in self.extra_returns_sections:
            self.register_returns_section(keyword, label, options)

        for keyword, label in self.extra_fields_sections:
            self.register_fields_section(keyword, label)

        for keyword, admonition in self.extra_version_info_sections:
            self.register_version_info_section(keyword, admonition)

        self._parse(True)

    def register_returns_section(self, keyword, label, options={}):
        """Register a Returns-like section with the given keyword and label.

        Args:
            keyword (unicode):
                The keyword used in the docs.

            label (unicode):
                The label outputted in the section.

            options (dict, optional):
                Options for the registration.

                This accepts:

                Keys:
                    require_type (bool, optional):
                        Whether the type is required, and assumed to be the
                        first line.

                Version Added:
                    2.0
        """
        self._sections[keyword] = lambda *args: \
            self._format_fields(
                label,
                self._consume_returns_section(**options))

    def register_fields_section(self, keyword, label):
        """Register a fields section with the given keyword and label.

        Args:
            keyword (unicode):
                The keyword used in the docs.

            label (unicode):
                The label outputted in the section.
        """
        self._sections[keyword] = lambda *args: \
            self._format_fields(label, self._consume_fields())

    def register_version_info_section(self, keyword, admonition):
        """Register a version section with the given keyword and admonition.

        Args:
            keyword (unicode):
                The keyword used in the docs.

            admonition (unicode):
                The admonition to use for the section.
        """
        self._sections[keyword] = lambda *args: \
            self._format_admonition_with_params(
                admonition,
                self._consume_to_next_section())

    def _format_admonition_with_params(self, admonition, lines):
        """Format an admonition section with the first line as a parameter.

        Args:
            admonition (unicode):
                The admonition name.

            lines (list of unicode):
                The list of lines to format.

        Returns:
            list of unicode:
            The resulting list of lines.
        """
        lines = self._strip_empty(lines)

        if lines:
            param_line = lines[0].strip()

            if param_line.endswith(':'):
                param_line = param_line[:-1]

            return (
                ['.. %s:: %s' % (admonition, param_line), ''] +
                self._indent(self._dedent(lines[1:]), 3) + ['']
            )
        else:
            return ['.. %s::' % admonition, '']

    def _parse(self, parse=False):
        """Parse the docstring.

        By default (when called from the parent class), this won't do anything.
        It requires passing ``True`` in order to parse. This is there to
        prevent the parent class from prematurely parsing before all sections
        are rendered.

        Args:
            parse (bool, optional):
                Set whether the parsing should actually happen.
        """
        if parse:
            super(BeanbagDocstring, self)._parse()

    def _consume_field(self, parse_type=True, *args, **kwargs):
        """Parse a field line and return the field's information.

        This enhances the default version from Napoleon to allow for attribute
        types that wrap across multiple lines.

        Args:
            parse_type (bool, optional):
                Whether to parse type information.

            *args (tuple):
                Position arguments to pass to the paren method.

            **kwargs (dict):
                Keyword arguments to pass to the paren method.

        Returns:
            tuple:
            Information on the field. The format is dependent on the parent
            method.
        """
        if parse_type:
            lines = self.peek_lines()
            m = self.partial_typed_arg_start_re.match(lines[0])

            if m:
                result = None

                for i in range(1, self.MAX_PARTIAL_TYPED_ARG_LINES):
                    # See if there's an ending part anywhere.
                    lines = self.peek_lines(i + 1)

                    if not isinstance(lines[i], str):
                        # We're past the strings and into something else.
                        # Bail.
                        break

                    m = self.partial_typed_arg_start_re.match(lines[i])

                    if m:
                        # We're in a new typed arg. Bail.
                        break

                    m = self.partial_typed_arg_end_re.match(lines[i])

                    if m:
                        # We need to rebuild the string based on the lines
                        # we've determined for the field. We have to be
                        # careful to join them in such a way where we have
                        # a space in-between if separating words, but not if
                        # separating parts of a class name.
                        parts = []
                        prev_line = None

                        for line in lines:
                            norm_line = line.strip()

                            if norm_line and prev_line is not None:
                                if (norm_line.startswith('.') or
                                    prev_line.endswith('.')):
                                    line = norm_line
                                else:
                                    parts.append(' ')

                            parts.append(line)
                            prev_line = norm_line

                        result = ''.join(parts)

                        if ',' in result:
                            result = ', '.join(self.COMMA_RE.split(result))

                        break

                if result:
                    # Consume those lines so they're not processed again.
                    self.consume_lines(len(lines))

                    # Insert the new resulting line in the line cache for
                    # processing.
                    self.queue_line(result)

        name, type_str, desc = super(BeanbagDocstring, self)._consume_field(
            parse_type, *args, **kwargs)

        return (name, self.make_type_reference(type_str), desc)

    def _consume_returns_section(self, *args, **kwargs):
        """Consume a returns section, converting and handling types.

        This enhances the default version from Napoleon to allow for
        return-like sections lacking a description, and to intelligently
        process type references.

        Version Added:
            2.0:
            This is the first release to override this functionality.

        Args:
            require_type (bool, optional):
                Whether to assume the first line is a type.

            *args (tuple):
                Position arguments to pass to the paren method.

            **kwargs (dict):
                Keyword arguments to pass to the paren method.

        Returns:
            tuple:
            Information on the field. The format is dependent on the parent
            method.
        """
        if kwargs.pop('require_type', False):
            lines = self.peek_lines(1)

            if lines:
                param_line = lines[0].rstrip()

                if not param_line.endswith(':'):
                    self.consume_lines(1)
                    self.queue_line('%s:' % param_line)

        nodes = super(BeanbagDocstring, self)._consume_returns_section(
            *args, **kwargs)

        return [
            (name, self.make_type_reference(type_str), desc)
            for name, type_str, desc in nodes
        ]

    def peek_lines(self, num_lines=1):
        """Return the specified number of lines without consuming them.

        Version Added:
            1.9

        Args:
            num_lines (int, optional):
                The number of lines to return.

        Returns:
            list of str:
            The resulting lines.
        """
        if self._USES_LINES_DEQUE:
            # Sphinx >= 5.1
            lines = self._lines

            return [
                lines.get(i)
                for i in range(num_lines)
            ]
        else:
            # Sphinx < 5.1
            return self._line_iter.peek(num_lines)

    def consume_lines(self, num_lines):
        """Consume the specified number of lines.

        This will ensure that these lines are not processed any further.

        Version Added:
            1.9

        Args:
            num_lines (int, optional):
                The number of lines to consume.
        """
        if self._USES_LINES_DEQUE:
            # Sphinx >= 5.1
            for i in range(num_lines):
                self._lines.popleft()
        else:
            # Sphinx < 5.1
            self._line_iter.next(num_lines)

    def queue_line(self, line):
        """Queue a line for processing.

        This will place the line at the beginning of the processing queue.

        Version Added:
            1.9

        Args:
            line (str):
                The line to queue.
        """
        if self._USES_LINES_DEQUE:
            # Sphinx >= 5.1
            self._lines.appendleft(line)
        else:
            # Sphinx < 5.1
            self._line_iter._cache.appendleft(line)

    def make_type_reference(self, type_str):
        """Create references to types in a type string.

        This will parse the string, separating out a type from zero or more
        suffixes (like ``, optional``).

        The type will be further parsed into a space-separated tokens. Each
        of those will be set as a reference, allowing Sphinx to try to link
        it. The exceptions are the words "of" and "or", which we use to
        list optional types.

        The suffixes are each formatted with emphasis.

        Version Added:
            2.0

        Args:
            type_str (unicode):
                The string to parse and create references from.

        Returns:
            unicode:
            The new string.
        """
        if not type_str:
            return type_str

        type_aliases = \
            getattr(self._config, 'napoleon_type_aliases', None) or {}
        parts = type_str.split(',')
        type_parts = []

        for type_part in parts[0].split(' '):
            if type_part not in ('of', 'or'):
                type_part = _convert_type_spec(type_part, type_aliases)

            type_parts.append(type_part)

        new_parts = [' '.join(type_parts)]

        if len(parts) > 1:
            new_parts += [
                r' *%s*' % _part.strip()
                for _part in parts[1:]
            ]

        return ','.join(new_parts)


def _filter_members(app, what, name, obj, skip, options):
    """Filter members out of the documentation.

    This will look up the name in the ``autodoc_excludes`` table under the
    ``what`` and under ``'*'`` keys. If an entry is listed, it will be
    excluded from the documentation.

    It will also skip modules listed in ``__deprecated__`` in the documented
    module.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application to register roles and configuration on.

        what (unicode):
            The type of thing being filtered (one of ``attribute``, ``class``,
            ``exception``, ``function``, ``method``, or ``module``).

        name (unicode):
            The name of the object to look up in the filter lists.

        obj (object):
            The object that may be filtered.

        skip (bool):
            Whether the filter will be skipped if this handler doesn't
            return a different value.

        options (object):
            Options given to the autodoc directive.

    Returns:
        bool:
        Whether the member will be skipped.
    """
    module_name = app.env.temp_data.get('autodoc:module')

    if not module_name:
        return

    module = sys.modules[module_name]

    # Check if the module itself is excluding this from the docs.
    module_excludes = getattr(module, '__autodoc_excludes__', [])

    if name in module_excludes:
        return True

    # Check if this appears in the global list of excludes.
    global_excludes = app.config['autodoc_excludes']

    for key in (what, '*'):
        if key in global_excludes and name in global_excludes.get(key, []):
            return True

    # Check if this appears in the list of deprecated objects.
    if name in getattr(module, '__deprecated__', []):
        return True

    return skip


def _process_docstring(app, what, name, obj, options, lines):
    """Process a docstring.

    If Beanbag docstrings are enabled, this will parse them and replace the
    docstring lines.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application.

        what (unicode):
            The type of thing owning the docstring.

        name (unicode):
            The name of the thing owning the docstring.

        obj (object):
            The object owning the docstring.

        options (dict):
            Options passed to this handler.

        lines (list of unicode):
            The lines to process.
    """
    if app.config['napoleon_beanbag_docstring']:
        docstring = BeanbagDocstring(lines, app.config, app, what, name, obj,
                                     options)
        lines[:] = docstring.lines()[:]


def _on_config_inited(app, config):
    """Override default configuration settings for Napoleon.

    This will ensure that some of our defaults take place for Beanbag
    docstring rendering.

    Version Added:
        2.0

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application to override configuration on.

        config (sphinx.config.Config):
            The Sphinx configuration to override.
    """
    if config.use_beanbag_autodoc_defaults:
        # Turn off competing docstring parsers.
        config.napoleon_google_docstring = False
        config.napoleon_numpy_docstring = False

        # Force this off so that we can handle this ourselves when consuming a
        # field.
        config.napoleon_preprocess_types = False

        # Change some defaults.
        config.values['autodoc_member_order'] = \
            ('bysource',) + config.values['autodoc_member_order'][1:]
        config.values['autoclass_content'] = \
            ('class',) + config.values['autoclass_content'][1:]

        config.autodoc_default_options = dict({
            'members': True,
            'special-members': True,
            'undoc-members': True,
            'show-inheritance': True,
        }, **(config.autodoc_default_options or {}))

    # Register type aliases.
    new_type_aliases = {}

    if 'python' in config.intersphinx_mapping:
        python_intersphinx = config.intersphinx_mapping['python'][0]

        if python_intersphinx is not None:
            if python_intersphinx.startswith('https://docs.python.org/2'):
                # Python 2
                #
                # Note that 'list' does not have corresponding type
                # documentation in Python 2.
                new_type_aliases.update({
                    'tuple': ':py:func:`tuple <tuple>`',
                    'unicode': ':py:func:`unicode <unicode>`',
                })
            else:
                # Python 3
                new_type_aliases.update({
                    'list': ':py:class:`list`',
                    'tuple': ':py:class:`tuple`',
                    'unicode': ':py:class:`unicode <str>`',
                })

    if new_type_aliases:
        if config.napoleon_type_aliases is None:
            config.napoleon_type_aliases = new_type_aliases.copy()
        else:
            config.napoleon_type_aliases.update(new_type_aliases)

        if 'autodoc_type_aliases' in config:
            config.autodoc_type_aliases.update(new_type_aliases)

    # Update autodoc_excludes to include defaults if requested.
    #
    # We'll also ensure all lists are sets, to make processing faster.
    if config.autodoc_excludes.get('__defaults__'):
        new_autodoc_excludes = {
            '*': {
                '__annotations__',
                '__dict__',
                '__doc__',
                '__module__',
                '__weakref__',
            },
        }

        if '*' in config.autodoc_excludes:
            new_autodoc_excludes['*'].update(config.autodoc_excludes['*'])

        new_autodoc_excludes.update({
            key: set(value)
            for key, value in config.autodoc_excludes.items()
            if key not in ('*', '__defaults__')
        })
    else:
        new_autodoc_excludes = {
            key: set(value)
            for key, value in config.autodoc_excludes.items()
            if key != '__defaults__'
        }

    config.autodoc_excludes = new_autodoc_excludes


def setup(app):
    """Set up the Sphinx extension.

    This sets up the configuration and event handlers needed for enhancing
    auto-doc support.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application to register configuration and listen to
            events on.
    """
    app.setup_extension('sphinx.ext.autodoc')
    app.setup_extension('sphinx.ext.intersphinx')
    app.setup_extension('sphinx.ext.napoleon')

    app.add_config_value('use_beanbag_autodoc_defaults', True, True)
    app.add_config_value('autodoc_excludes', {'__defaults__': True}, True)
    app.add_config_value('napoleon_beanbag_docstring', True, True)

    app.connect('autodoc-skip-member', _filter_members)
    app.connect('autodoc-process-docstring', _process_docstring)
    app.connect('config-inited', _on_config_inited)

    return {
        'version': VERSION,
        'parallel_read_safe': True,
    }



"""Sphinx extension to improve references.

This enhances references, allowing both Python and JavaScript references to
break paths (like ``foo.bar.MyClass``) across multiple lines.


Setup
=====

To use this, add the extension in :file:`conf.py`::

    extensions = [
        ...
        'beanbag_docutils.sphinx.ext.ref_utils',
        ...
    ]
"""

import re

from sphinx.domains.javascript import JSXRefRole
from sphinx.domains.python import PyXRefRole


WS_RE = re.compile(r'\s+')


def _process_link(orig_process_link, role, *args, **kwargs):
    """Process a link, stripping away all whitespace from targets.

    Args:
        orig_process_link (callable):
            The original function used to process a link for the role.

        role (sphinx.roles.XRefRole):
            The role instance being processed.

        *args (tuple):
            Positional arguments to pass to the original function.

        **kwargs (dict):
            Keyword arguments to pass to the original function.

    Returns:
        tuple:
        A tuple of ``(title, target)``.
    """
    title, target = orig_process_link(role, *args, **kwargs)

    return title, WS_RE.sub('', target)


def setup(app):
    """Set up the Sphinx extension.

    This alters some behavior for the Python domain, allowing newlines in
    references.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application to extend.
    """
    def _setup_ref_role(role):
        orig_process_link = role.process_link
        role.process_link = \
            lambda *args, **kwargs: \
            _process_link(orig_process_link, *args, **kwargs)

    for role in (JSXRefRole, PyXRefRole):
        _setup_ref_role(role)
